#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "lik.h"
#include "triplet.h"
#include "calclik.h"

/* prototypes */

TNode *startNode;
double (*CalcLik)();

void CalcTripletLikAmbiguous(int branchOfs);
void CalcTripletLik(int branchOfs);


#define Ra R[0]
#define Rc R[1]
#define Rg R[2]
#define Rt R[3]

static double logNumCats;


/**********************************/	
void SetupLikelihood(TTree *tree)
{
	/* pre-calculate the log of the number of gamma categories */
	logNumCats=log(numCategories);
	InitFunc(globalMLE, tree->nodes);
	if (gammaRates)
		CalcLik=gammaCalcLik;
	else
		CalcLik=codonCalcLik;
}


/*************************************/
double codonCalcLik(int region, int pntRecomb1, int pntRecomb2)
{
	int k;
	double sum;
	short *P, *Q;

	sum=0.0;
	if (!region) {
		if (ambiguousBases)
			CalcTripletLikAmbiguous(0);
		else
			CalcTripletLik(0);
			
		P=categories;
		Q=patternIndex;
		for(k=0; k<pntRecomb1; k++) {			
			sum+=log(patternLik[*P][*Q]);
			P++;
			Q++;
		}
		
		P=categories+pntRecomb2;
		Q=patternIndex+pntRecomb2;
		for(k=pntRecomb2; k<numBases; k++) {			
			sum+=log(patternLik[*P][*Q]);
			P++;
			Q++;
		}
	} else {
		if (ambiguousBases)
			CalcTripletLikAmbiguous(3);
		else
			CalcTripletLik(3);
			
		P=categories+pntRecomb1;
		Q=patternIndex+pntRecomb1;
		for(k=pntRecomb1; k<pntRecomb2; k++) {			
			sum+=log(patternLik[*P][*Q]);
			P++;
			Q++;
		}
	}
		
	return sum;
}


/*************************************/
double gammaCalcLik(int region, int pntRecomb1, int pntRecomb2)
{
	int cat, k;
	double sum;
	double *P, *Q;
	short *T;

	sum=0.0;
	
	if (!region) {
		if (ambiguousBases)
			CalcTripletLikAmbiguous(0);
		else
			CalcTripletLik(0);

		/* sum across gamma rate categories */		
		for (cat=1; cat<numCategories; cat++) {
			P=patternLik[0];
			Q=patternLik[cat];
			for(k=0; k<numPatterns; k++) {			
				*P+=*Q;
				P++;
				Q++;
			}
		}
	
		/* then multiply the lot together (i.e. sum the logs) */
		T=patternIndex;
		for(k=0; k<pntRecomb1; k++) {			
			sum+=(log(patternLik[0][*T])-logNumCats);
			T++;
		}
	
		/* and now the second region */
		T=patternIndex+pntRecomb2;
		for(k=pntRecomb2; k<numBases; k++) {			
			sum+=(log(patternLik[0][*T])-logNumCats);
			T++;
		}
	
	} else {
		if (ambiguousBases)
			CalcTripletLikAmbiguous(3);
		else
			CalcTripletLik(3);
		
		/* sum across gamma rate categories */		
		for (cat=1; cat<numCategories; cat++) {
			P=patternLik[0];
			Q=patternLik[cat];
			for(k=0; k<numPatterns; k++) {			
				*P+=*Q;
				P++;
				Q++;
			}
		}
	
		/* then multiply the lot together (i.e. sum the logs) */
		T=patternIndex+pntRecomb1;
		for(k=pntRecomb1; k<pntRecomb2; k++) {			
			sum+=(log(patternLik[0][*T])-logNumCats);
			T++;
		}
	}
		
	return sum;
}


/*************************************/
void CalcTripletLikAmbiguous(int branchOfs)
{
	int k, kA, kC, kG, kT, n, cat;
	double *P, *Q, *R, *U, *lut;
	TNode *newposn;

	n=numPatterns;

	for (cat=0; cat<numCategories; cat++) {
		newposn=startNode->next->branch;		
		P=startNode->Likelihood;
		Q=newposn->Likelihood;
		lut=transMatrix[cat]+((newposn->branchNo+branchOfs) << 4);

		for (k=0; k<n; k++) {
			kA=k*4; kC=kA+1; kG=kC+1; kT=kG+1;
			P[kA]=lut[0]*Q[kA] + lut[1]*Q[kC] + lut[2]*Q[kG] + lut[3]*Q[kT];
			P[kC]=lut[4]*Q[kA] + lut[5]*Q[kC] + lut[6]*Q[kG] + lut[7]*Q[kT];
			P[kG]=lut[8]*Q[kA] + lut[9]*Q[kC] + lut[10]*Q[kG] + lut[11]*Q[kT];
			P[kT]=lut[12]*Q[kA] + lut[13]*Q[kC] + lut[14]*Q[kG] + lut[15]*Q[kT];
		}
		
		newposn=startNode->next->next->branch;
		Q=newposn->Likelihood;
		lut=transMatrix[cat]+((newposn->branchNo+branchOfs) << 4);
		
		for (k=0; k<n; k++) {
			kA=k*4; kC=kA+1; kG=kC+1; kT=kG+1;
			P[kA]*=lut[0]*Q[kA] + lut[1]*Q[kC] + lut[2]*Q[kG] + lut[3]*Q[kT];
			P[kC]*=lut[4]*Q[kA] + lut[5]*Q[kC] + lut[6]*Q[kG] + lut[7]*Q[kT];
			P[kG]*=lut[8]*Q[kA] + lut[9]*Q[kC] + lut[10]*Q[kG] + lut[11]*Q[kT];
			P[kT]*=lut[12]*Q[kA] + lut[13]*Q[kC] + lut[14]*Q[kG] + lut[15]*Q[kT];
		}

		lut=transMatrix[cat]+((startNode->branchNo+branchOfs) << 4);
		P=patternLik[cat];
		Q=startNode->Likelihood;
		U=startNode->branch->Likelihood;

		for (k=0; k<n; k++) {
			kA=k*4; kC=kA+1; kG=kC+1; kT=kG+1;
			P[k]=freqA*U[kA]*(lut[0]*Q[kA] + lut[1]*Q[kC] + lut[2]*Q[kG] + lut[3]*Q[kT]) +
				freqC*U[kC]*(lut[4]*Q[kA] + lut[5]*Q[kC] + lut[6]*Q[kG] + lut[7]*Q[kT]) +
				freqG*U[kG]*(lut[8]*Q[kA] + lut[9]*Q[kC] + lut[10]*Q[kG] + lut[11]*Q[kT]) +
				freqT*U[kT]*(lut[12]*Q[kA] + lut[13]*Q[kC] + lut[14]*Q[kG] + lut[15]*Q[kT]);
		}
	}
}


/*************************************/
void CalcTripletLik(int branchOfs)
{
	int cat, k;
	double *P, *Q, *lut, PA, PC, PG, PT;
	char *S;
	
	for (cat=0; cat<numCategories; cat++) {
		P=patternLik[cat];
				
		lut=transMatrix[cat]+(branchOfs << 4);

		for (k=0; k<numPatterns; k++) {
			S=patterns[k];
			Q=lut;
			switch (*S) {
				case 'A': 
					PA=Q[0];
					PC=Q[4];
					PG=Q[8];
					PT=Q[12];
				break;
				case 'C': 
					PA=Q[1];
					PC=Q[5];
					PG=Q[9];
					PT=Q[13];
				break;
				case 'G': 
					PA=Q[2];
					PC=Q[6];
					PG=Q[10];
					PT=Q[14];
				break;
				case 'T': 
					PA=Q[3];
					PC=Q[7];
					PG=Q[11];
					PT=Q[15];
				break;
			}
			S++;
			Q+=16;
			switch (*S) {
				case 'A': 
					PA*=Q[0];
					PC*=Q[4];
					PG*=Q[8];
					PT*=Q[12];
				break;
				case 'C': 
					PA*=Q[1];
					PC*=Q[5];
					PG*=Q[9];
					PT*=Q[13];
				break;
				case 'G': 
					PA*=Q[2];
					PC*=Q[6];
					PG*=Q[10];
					PT*=Q[14];
				break;
				case 'T': 
					PA*=Q[3];
					PC*=Q[7];
					PG*=Q[11];
					PT*=Q[15];
				break;
			}			
			S++;
			Q+=16;
			switch (*S) {
				case 'A': 
					PA*=freqA*Q[0];
					PC*=freqC*Q[4];
					PG*=freqG*Q[8];
					PT*=freqT*Q[12];
				break;
				case 'C': 
					PA*=freqA*Q[1];
					PC*=freqC*Q[5];
					PG*=freqG*Q[9];
					PT*=freqT*Q[13];
				break;
				case 'G': 
					PA*=freqA*Q[2];
					PC*=freqC*Q[6];
					PG*=freqG*Q[10];
					PT*=freqT*Q[14];
				break;
				case 'T': 
					PA*=freqA*Q[3];
					PC*=freqC*Q[7];
					PG*=freqG*Q[11];
					PT*=freqT*Q[15];
				break;
			}
			*P=PA+PC+PG+PT;
			P++;
		}
	}
}



