/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly  */
/*  Department of Zoology, University of Oxford        */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "lik.h"
#include "triplet.h"


/* prototypes */

void ConnectBranch(TNode *destNode, TNode *srcNode, int b, TTree *tree);

/* functions */

TTree *CreateTriplet()
{
	TTree *tree;
	TNode *P;
	int i;
	
	tree=(TTree *)AllocMem(sizeof(TTree), "tree", "CreateTree", 1);

	tree->numTips=3;
	tree->numNodes=1;
	tree->totalNodes=numTaxa+(3*tree->numNodes);

	tree->nodes=(TNode *)AllocMem(sizeof(TNode)*tree->totalNodes, "tree->nodes", "CreateTree", 1);

	P=tree->nodes;
	for (i=0; i<tree->numTips; i++) {
		P->branch=NULL;
		P->next=NULL;
		
		P->Likelihood=NULL;	
		P->tipNo=i;
		P->nodeNo=-1;
	
		P++;
	}	
	
	tree->start=P;
	
	for (i=0; i<tree->numNodes; i++) {
		
		P->Likelihood=NULL;
		P->branch=NULL;
		P->tipNo=-1;
		P->nodeNo=i;
		P->next=P+1;
		
		(P+1)->branch=NULL;
		(P+1)->Likelihood=NULL;
		(P+1)->tipNo=-1;
		(P+1)->nodeNo=i;
		(P+1)->next=P+2;
		
		(P+2)->branch=NULL;
		(P+2)->Likelihood=NULL;
		(P+2)->tipNo=-1;
		(P+2)->nodeNo=i;
		(P+2)->next=P;

		P+=3;
	}
	
	tree->branches=(TNode **)AllocMem(sizeof(TNode *)*numBranches, "tree->branches", "CreateTree", 1);
	
	return tree;
}


void DisposeTriplet(TTree *tree)
{
	int i;
	
	if (tree!=NULL) {
		for (i=0; i<tree->totalNodes; i++)
			free(tree->nodes[i].Likelihood);
		
		if (tree->nodes!=NULL)
			free(tree->nodes);
		
		if (tree->branches!=NULL)
			free(tree->branches);
		
		free(tree);
	}
}


void ConnectBranch(TNode *destNode, TNode *srcNode, int b, TTree *tree)
{
	srcNode->branch=destNode;
	destNode->branch=srcNode;
	destNode->branchNo=srcNode->branchNo=b;
	tree->branches[b]=destNode;
}	


void SetTriplet(TTree *tree)
{
	TNode *node;

	node=tree->nodes+tree->numTips;
	
	ConnectBranch(tree->nodes, node, 0, tree);
	ConnectBranch(tree->nodes+1, node->next, 1, tree);
	ConnectBranch(tree->nodes+2, node->next->next, 2, tree);
	
	tree->start=node;
}

